#!/usr/bin/perl -w
#
# Generate raceprot.c from pre.c + automatically generated code + post.c
#


my $rootdir = ".";


#
# Extract OpenBSD syscall names and their numbers
#

open F, "/sys/kern/init_sysent.c" or die "Cannot open /sys/kern/init_sysent.c: $!";

while (<F>)
{
	if (/(\d+) = (.*) \*/)
	{
		$syscall[$1] = $2;
		$syscall{$2} = $1;
	}
}

close F;


#
# Find out what system calls we are interested in
#

chdir($rootdir) or die "Cannot chdir($rootdir): $!\n";

open F, "syscalls" or die $!;

while(<F>)
{
	chomp;
	next if /^\s*$/;		# skip blank lines
	next if /^;/;			# skip comments

	if (/^(\w+)$/)
	{
		$s{$1} = 1;
	}
}

close F;


#
# Create raceprot.c
#

open RPC, ">raceprot.c" or die "Cannot create raceprot.c: $!";

print RPC "/* DO NOT EDIT. AUTOMATICALLY GENERATED FILE. SEE README & generate.pl */\n\n";


#
# Prepend the contents of pre.c file to raceprot.c
#

open F, "pre.c" or die "Cannot open pre.c: $!";
print RPC <F>;
close F;


# the code for this functions is not automatically generated (see post.c)

my @builtin = qw(exit fork rfork vfork execve);
#my @builtin = qw(exit fork rfork vfork close execve);


#
# Generate function prototypes
#

my $i;

for (@builtin)
{
	print RPC "int raceprot_sys_$_(struct proc*, void*, register_t*);\n";
}

for ($i = 0; $i < @syscall; $i++)
{
	if (exists $s{$syscall[$i]})
	{
		print RPC "int raceprot_sys_$syscall[$i](struct proc*, void*, register_t*);\n";
	}
}


print RPC "\n\n";



#
# Table of system calls to intercept
#

print RPC "struct backup_sysent backup_sysent[] = {\n";

my $j = 0;

for (@builtin)
{
	print RPC "#define NUM_SYS_$_ $j\n";
	print RPC "\tBACKUP_SYSENT($syscall{$_}, $_)\n";

	++$j;
}

for ($i = 0; $i < @syscall; $i++)
{
	if (exists $s{$syscall[$i]})
	{
		print RPC "#define NUM_SYS_$syscall[$i] $j\n";
		print RPC "\tBACKUP_SYSENT($i, $syscall[$i])\n";

		++$j;
	}
}

print RPC "};\n\n";



#
# Create raceprot.h
#

open RPH, ">raceprot.h" or die "Cannot open raceprot.h: $!";

print RPH "/* DO NOT EDIT. AUTOMATICALLY GENERATED FILE. SEE README & generate.pl */\n\n";


#
# Prepend the contents of pre.h file to raceprot.h
#

open F, "pre.h" or die "Cannot open pre.h: $!";
print RPH <F>;
close F;


#
# Generate enum FILEOP (enumeration of all the file operations that we mediate)
#

print RPH "enum FILEOP {NONE=0, OPEN_READ=1, OPEN_WRITE=2, OPEN_CREAT=4, RENAME_FROM=8, RENAME_TO=16, CLOSE=32, EXEC=64, STAT=128, LSTAT=256, ";

$j = 512;		# next operation index after LSTAT=256

my $code = "";

my @pregenerated;

# code for this system calls can be found in post.c
$pregenerated[$syscall{rename}] = $pregenerated[$syscall{stat}] =
$pregenerated[$syscall{lstat}] = 1;


for ($i = 6; $i < @syscall; $i++)
{
	next if $pregenerated[$i];

	if (exists $s{$syscall[$i]})
	{
		# enum FILEOP entry

		print RPH qq!\U$syscall[$i]=$j, !;
		$j = $j * 2;


		# code to convert enum fileop entry to a string

		$code .= qq!\tif (operation & \U$syscall[$i]\E)\t\treturn "$syscall[$i]";\n!;
	}
}

print RPH "};";

print RPC qq!


char*
fileop_name(enum FILEOP operation)
{
	if (operation & OPEN_READ)		return "ropen";
	if (operation & OPEN_WRITE)	return "wopen";
	if (operation & OPEN_CREAT)	return "copen";
	if (operation & RENAME_FROM)	return "frename";
	if (operation & RENAME_TO)		return "trename";
	if (operation & STAT)		return "stat";
	if (operation & LSTAT)		return "lstat";
	if (operation & CLOSE)		return "close";
	if (operation & EXEC)		return "exec";
$code
	return "none";
}
!;



#
# Generate actual code
#

my ($arg, $addcode, $op);

# $i = 6: skip open() call - special case

for ($i = 6; $i < @syscall; $i++)
{
	next if $pregenerated[$i];

	$arg = "syscallarg(char *) path;";
	$addcode = "";
	$op = "RP_OP_NORMAL";

	if ($i == $syscall{"link"} || $i == $syscall{"symlink"})
	{
		# link & symlink need to process the second char* parameter

		$arg = "syscallarg(char *) path2;\n";
		$arg .= "\t\tsyscallarg(char *) path;";
	}
	elsif ($i == $syscall{"access"})
	{
		# access call requires extra code to handle the non-existent
		# file case

		$addcode = qq!\n\t\t/* save the information about the non-existent file */
		rp_insert(p->p_pid, 0, 0, (char*) SCARG(uap, path), ACCESS, ticks, RP_OP_STAT);\n!;
	}
	elsif ($i == $syscall{"unlink"})
	{
		$op = "RP_OP_UNLINK";
	}


	if (exists $s{$syscall[$i]})
	{
		print RPC qq@
/*
 * Mediate sys_$syscall[$i] system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_$syscall[$i](struct proc *p, void *v, register_t *retval)
{
	struct sys_$syscall[$i]_args /* {
		$arg
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_$syscall[$i](%s)\\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{$addcode
		return CALL_SAVED_SYS_CALL($syscall[$i]);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), \U$syscall[$i]\E, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, \U$syscall[$i]\E, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL($syscall[$i]);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), \U$syscall[$i]\E, ticks, $op);
	}

	return ret2;
}


@;
	}
}



#
# Append the contents of post.c file to raceprot.c
#

open F, "post.c" or die "Cannot open post.c: $!";
print RPC <F>;
close F;


close RPC;


#
# Append the contents of post.h file to raceprot.h
#

open F, "post.h" or die "Cannot open post.h: $!";
print RPH <F>;
close F;


close RPH;


