

/*
 * raceprot_respond:	Respond to a possible race condition attack:
 *			either suspend a process (depending on a suspend var)
 *			or kill it
 *
 * input:		struct proc *p		used for sending a kill signal
 *			enum FILEOP operation	operation that caused the
 *						race condition
 *			const char *path	name of a file object that is
 *						involved
 *			int obj_type		type (directory / file) of a
 *						file object that caused the
 *						race condition
 */

void
raceprot_respond(struct proc *p, enum FILEOP operation, const char *path, int obj_type)
{
	if (suspend)
	{
		int	timeout;

		timeout = (obj_type == VDIR ? TIMEOUT_DIR : TIMEOUT_FILE);
		timeout += loadavg;

		log(9, "Potentially dangerous operation %d %s(%s). Suspending process %d (%s) for %d seconds\n", operation, fileop_name(operation), path, p->p_pid, p->p_comm, timeout / 100);

		tsleep("susp", PZERO, "raceprot", timeout);
		//tsleep("susp", PPAUSE, "raceprot", timeout);
	}
	else
	{
		log(9, "Potentially dangerous operation %s(%s). Killing process %d (%s)\n", fileop_name(operation), path, p->p_pid, p->p_comm);

		psignal(p, 9);
	}
}



/*
 * Mediate sys_stat system call:
 *
 * Call the original stat system call.
 *
 * We save information about non-existent files
 * to detect temporary file race conditions.
 */

int
raceprot_sys_stat(struct proc *p, void *v, register_t *retval)
{
        struct sys_stat_args /* {
                syscallarg(char *) path;
        } */ *uap = v;
        int                     ret;


	ret = CALL_SAVED_SYS_CALL(stat);


	if (verbose)
		uprintf("%d sys_stat(%s) = %d\n", p->p_pid, SCARG(uap, path), ret);


	/* did stat fail? */
	if (ret == ENOENT)
	{
		/* save information about the non-existent file */

		rp_insert(p->p_pid, 0, 0, (char*) SCARG(uap, path), STAT,
			  ticks, RP_OP_STAT);
	}

	return ret;
}



/*
 * Mediate sys_lstat system call:
 *
 * Call the original stat system call.
 *
 * We save information about non-existent files
 * to detect temporary file race conditions.
 */

int
raceprot_sys_lstat(struct proc *p, void *v, register_t *retval)
{
        struct sys_lstat_args /* {
                syscallarg(char *) path;
        } */ *uap = v;
        int                     ret;


	ret = CALL_SAVED_SYS_CALL(lstat);


	if (verbose)
		uprintf("%d sys_lstat(%s) = %d\n", p->p_pid, SCARG(uap, path), ret);


	/* did lstat fail? */
	if (ret == ENOENT)
	{
		/* save information about the non-existent file */

		rp_insert(p->p_pid, 0, 0, (char*) SCARG(uap, path), LSTAT,
			  ticks, RP_OP_STAT);
	}

	return ret;
}



/*
 * raceprot_sys_open:	Mediate sys_open call
 *
 * Note: there is a race condition between this function and sys_open:
 *	 a filename can be changed in the userspace before sys_open gets to
 *	 process it (reference: LOMAC you can live with)
 */

int
raceprot_sys_open(struct proc *p, void *v, register_t *retval)
{
	struct sys_open_args /* {
		syscallarg(char *) path;
		syscallarg(int) flags;
		syscallarg(int) mode;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2, op;
	struct vattr		va;


	if (verbose)
		uprintf("%d (%s) sys_open(%s)\n", p->p_pid, p->p_comm,
							SCARG(uap, path));

	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file exist? */

	if (namei(&nd) == 0)
	{
		/* get the inode */

		VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);

		op = SCARG(uap, flags) & (O_RDWR | O_WRONLY | O_CREAT) ? OPEN_WRITE : OPEN_READ;

		/* special case, insecure O_CREAT */
		if (SCARG(uap, flags) & (O_CREAT & ~O_EXCL))
			op = op | OPEN_CREAT;


		ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
					(char*) SCARG(uap, path), op, 1);


		if (ret == RP_PROCESS_RESPOND)
		{
			raceprot_respond(p, op, SCARG(uap, path),
							nd.ni_vp->v_type);
		}
		else if (ret == RP_PROCESS_FAIL_CALL)
		{
			/* fail the insecure O_CREAT. force the O_EXCL behavior */

			if (verbose)
				uprintf("insecure O_CREAT detected\n");

			return(EEXIST);
		}


		ret2 = CALL_SAVED_SYS_CALL(open);

		if (ret != RP_PROCESS_OK && ret2 == 0)
		    rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			      (char*)SCARG(uap, path), op, ticks, RP_OP_NORMAL);


		if (verbose)
			uprintf("%d sys_open(%s, %d) = %d (fd %d)\n", p->p_pid, SCARG(uap, path), op, ret, retval[0]);


#ifdef MEDIATE_SYS_CLOSE
		if (ret2 == 0 && retval[0] != -1)
			rp_register_fd(p->p_pid, va.va_fileid, retval[0]);
#endif

		vrele(nd.ni_vp);

		return(ret2);
	}


	/* CASE 2: file does not exit. open might create it */

	ret2 = CALL_SAVED_SYS_CALL(open);

	ret = namei(&nd);

	if (verbose)
		uprintf("%d (2) sys_open(%s) = %d (fd %d)\n", p->p_pid, SCARG(uap, path), ret2, retval[0]);


	if (ret == 0)
	{
		/* get the inode */

		VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);

		op = SCARG(uap, flags) & (O_RDWR | O_WRONLY | O_CREAT) ? OPEN_WRITE : OPEN_READ;

		/* special case, insecure O_CREAT */
		if (SCARG(uap, flags) & (O_CREAT & ~O_EXCL))
			op = op | OPEN_CREAT;


		ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
					(char*) SCARG(uap, path), op, 0);


		if (ret == RP_PROCESS_RESPOND)
		{
			/* too late to respond since the system
				call has finished by now! */

			uprintf("raceprot: sys_open() finished already (%d %s, %s)\n", op, SCARG(uap, path), p->p_comm);
/*
			raceprot_respond(p, op, SCARG(uap, path), nd.ni_vp->v_type);
*/
		}
		else if (ret == RP_PROCESS_FAIL_CALL)
		{
			uprintf("raceprot: fail call sys_open() finished already (%d %s)\n", op, SCARG(uap, path));
		}


		if (ret != RP_PROCESS_OK && ret2 == 0)
		    rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			      (char*)SCARG(uap, path), op, ticks, RP_OP_NORMAL);


#ifdef MEDIATE_SYS_CLOSE
		if (retval[0] != -1)
			rp_register_fd(p->p_pid, va.va_fileid, retval[0]);
#endif

	}

	if (nd.ni_vp != NULL)
		vrele(nd.ni_vp);

	return(ret2);
}



/*
 * Mediate sys_execve system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_execve(struct proc *p, void *v, register_t *retval)
{
	struct sys_execve_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d (%s) sys_execve(%s)\n", p->p_pid, p->p_comm, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(execve);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), EXEC, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, EXEC, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	if (ret != RP_PROCESS_OK)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
		          (char*) SCARG(uap, path), EXEC, ticks, RP_OP_NORMAL);
	}


	ret2 = CALL_SAVED_SYS_CALL(execve);

		
	/* we don't trust the exec'ed process thus we remove all the entries
	   belonging to the original process */

	if (ret2 == 0)
		rp_purge(PURGE_TYPE_PID, p->p_pid, 0);

	return(ret2);
}



/*
 * Mediate sys_rename system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_rename(struct proc *p, void *v, register_t *retval)
{
	struct sys_rename_args /* {
		syscallarg(char *) from;
		syscallarg(char *) to;
	} */ *uap = v;
	struct nameidata	nd, nd2;
	int			ret, ret2, ret3;
	struct vattr		va;


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, from), p);
	NDINIT(&nd2, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, to), p);

	ret = namei(&nd);
	ret2 = namei(&nd2);

	if (verbose)
		uprintf("%d (%s) sys_rename(%s, %s) = %d %d\n", p->p_pid, p->p_comm, SCARG(uap, from), SCARG(uap, to), ret, ret2);


	/* does file exist? */

	if (ret == 0)
	{
		/* get the inode */

		VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


		ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
				 (char*) SCARG(uap, from), RENAME_FROM, 0);

		if (ret == RP_PROCESS_RESPOND)
		{
			raceprot_respond(p, RENAME_FROM, SCARG(uap, from),
							 nd.ni_vp->v_type);
		}

		vrele(nd.ni_vp);


		if (ret2 == 0)
		{
			/* get the inode */

			VOP_GETATTR(nd2.ni_vp, &va, p->p_ucred, p);


			ret2 = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
				(char*) SCARG(uap, to), RENAME_TO, 0);

			if (ret2 == RP_PROCESS_RESPOND)
			{
				raceprot_respond(p, RENAME_TO, SCARG(uap, to), nd2.ni_vp->v_type);
			}

			vrele(nd2.ni_vp);
		}
	}


	ret3 = CALL_SAVED_SYS_CALL(rename);

	if (ret != RP_PROCESS_OK && ret2 != RP_PROCESS_OK && ret3 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			  (char*) SCARG(uap, to), RENAME_TO, ticks, RP_OP_NORMAL);
	}

	return ret3;
}



/*
 * Mediate sys_close system call:
 * purge the entry belonging to the current process with the file descriptor fd
 */

#ifdef MEDIATE_SYS_CLOSE

int
raceprot_sys_close(struct proc *p, void *v, register_t *retval)
{
	struct sys_close_args	*uap = v;
	int			ret;


	ret = CALL_SAVED_SYS_CALL(close);

	if (verbose)
		uprintf("%d sys_close(%d) %d\n", p->p_pid, SCARG(uap, fd), ret);

	if (ret == 0)
		rp_purge(PURGE_TYPE_FD, p->p_pid, SCARG(uap, fd));


	return(ret);
}

#endif



/*
 * Mediate sys_fork system call:
 * duplicate all the entries belonging to the current process with the new pid
 */

int
raceprot_sys_fork(struct proc *p, void *v, register_t *retval)
{
	int			ret;


	ret = CALL_SAVED_SYS_CALL(fork);

	if (verbose)
		uprintf("%d sys_fork() %d\n", p->p_pid, retval[0]);

	if (ret == 0)
		rp_duplicate_entry(p->p_pid, retval[0]);


	return(ret);
}


/*
 * Mediate sys_vfork system call:
 * duplicate all the entries belonging to the current process with the new pid
 */

int
raceprot_sys_vfork(struct proc *p, void *v, register_t *retval)
{
	int			ret;


	ret = CALL_SAVED_SYS_CALL(vfork);

	if (verbose)
		uprintf("%d sys_vfork() %d\n", p->p_pid, retval[0]);

	if (ret == 0)
		rp_duplicate_entry(p->p_pid, retval[0]);


	return(ret);
}



/*
 * Mediate sys_rfork system call:
 * duplicate all the entries belonging to the current process with the new pid
 */

int
raceprot_sys_rfork(struct proc *p, void *v, register_t *retval)
{
	int			ret;


	ret = CALL_SAVED_SYS_CALL(rfork);

	if (verbose)
		uprintf("%d sys_rfork() %d\n", p->p_pid, retval[0]);

	if (ret == 0)
		rp_duplicate_entry(p->p_pid, retval[0]);


	return(ret);
}



/*
 * Mediate sys_exit system call:
 * purge all the entries belonging to the terminated process
 */

int
raceprot_sys_exit(struct proc *p, void *v, register_t *retval)
{
	if (verbose)
		uprintf("%d sys_exit()\n", p->p_pid);

	rp_purge(PURGE_TYPE_PID, p->p_pid, 0);

	return CALL_SAVED_SYS_CALL(exit);
}



/*
 * Kernel Module Handler
 * Load / Unload the module
 */

static int
raceprot_handle(struct lkm_table *lkmt, int cmd)
{
	int		i, num;


	switch(cmd) {
	case LKM_E_LOAD:
		/* don't load twice! */

		if (lkmexists(lkmt))
			return (EEXIST);


		/* initialize the race protection system */

		rp_init();


		/* intercept the system calls listed in backup_sysent table */

		for (i = 0; i < BACKUP_SYSENT_NUM_ENTRIES; i++)
		{
			num = backup_sysent[i].sys_number;

			if (verbose)
				uprintf("loading %s(%d)\n",
					backup_sysent[i].sys_name, num);

			backup_sysent[i].old_sys_call = sysent[num].sy_call;
			sysent[num].sy_call = backup_sysent[i].new_sys_call;
		}

		break;


	case LKM_E_UNLOAD:

		/* restore the system calls listed in backup_sysent table */

		for (i = 0; i < BACKUP_SYSENT_NUM_ENTRIES; i++)
		{
			num = backup_sysent[i].sys_number;

			if (verbose)
				uprintf("unloading %s(%d)\n",
					backup_sysent[i].sys_name, num);

			sysent[num].sy_call = backup_sysent[i].old_sys_call;
		}

		rp_fini();


		break;
	}

	return(0);
}



/*
 * Kernel Module Entry Point
 */

int
raceprot(struct lkm_table *lkmt, int cmd, int ver)
{
	DISPATCH(lkmt, cmd, ver, raceprot_handle, raceprot_handle, lkm_nofunc)
}


