/*
 * Copyright (c) 2002-2004 Eugene Tsyrklevich (eugene@securityarchitects.com)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/mount.h>
#include <sys/exec.h>
#include <sys/lkm.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/syscall.h>
#include <sys/syscallargs.h>
#include <sys/proc.h>
#include <sys/namei.h>
#include <sys/vnode.h>
#include <sys/malloc.h>
#include <sys/timeout.h>
#include <sys/signalvar.h>
#include <sys/syslog.h>

#include "raceprot.h"
#include "lock.h"


MOD_MISC("raceprot")


int 		verbose = 0;
int 		suspend = 1;	/* suspend processes instead of killing them */
int		deny_all = 1;	/* 1 - deny_all, 0 - permit_all policy */


int		unloading_module, loadavg, timeout_purge, purge_needed;

struct timeout	to;

struct rp_entry	free_list = {NULL, 0, 0, 0, 0, 0, NONE};
int		free_list_num = 0;



#define	rp_HASH_SIZE		997
struct rp_entry	*rp_hash[rp_HASH_SIZE];

#define	stat_HASH_SIZE		7
struct rp_entry	*stat_hash[stat_HASH_SIZE];

#define	unlink_HASH_SIZE	7
struct rp_entry	*unlink_hash[unlink_HASH_SIZE];


#define	HASH_FIRST(name, value)	(name##_hash[value % name##_HASH_SIZE])





/*
 * rp_process:	Verify that the current operation does not interfere with
 *		any previous operations involving the same inode
 *
 * input:	pid of the process invoking the operation
 *		inode involved in the operation
 *		id of a filesystem where inode is residing
 *		[file]name associated with the inode
 *		operation being carried on the inode
 *		process_special_cases is set to 1 if 2 special OPEN_CREAT cases
 *			need to be checked (used only by raceprot_sys_open)
 *
 * output:	RP_PROCESS_OK		system call should be allowed to proceed
 *		RP_PROCESS_INSERT	system call should be allowed to proceed
 *					and information about the operation
 *					should be saved (rp_insert())
 *		RP_PROCESS_RESPOND	process should be suspended/killed
 *		RP_PROCESS_FAIL_CALL	system call should be aborted
 */

int
rp_process(pid_t pid, int inode, int fsid, char *name, enum FILEOP operation,
	   int process_special_cases)
{
	struct rp_entry	*tmp;
	int		found = 0;


	/* purge_needed flag is set by the callout procedure every second */

	if (purge_needed)
	{
		/*
		 * Dynamically calculate the timeout used to determine when
		 * to purge operation entries. The higher system load average
		 * is, the longer we keep the entries.
		 */

		loadavg = (averunnable.ldavg[0] * 100 + FSCALE / 2) >> FSHIFT;

		timeout_purge = min(TIMEOUT_PURGE + loadavg, TIMEOUT_PURGE_MAX);

		rp_purge(PURGE_TYPE_TIMEOUT, 0, 0);
		purge_needed = 0;
	}


	if (lock_shared() != 0)
	{
		uprintf("rp_process: failed to obtain a shared lock\n");
		return(0);
	}


	/* deal with 2 special cases */

	if (process_special_cases && (operation & OPEN_CREAT))
	{
		/*
		 * special case 1:
	 	 *
		 * temporary file race / [l]stat/access followed by open:
		 *	Victim:		calls [l]stat/access on a file &
		 *			file does not exist
		 *	Attacker:	creates a link with the same name
		 *	Victim:		open(O_CREAT, name) opens the existing
		 * 			file and trashes its contents
		 *
		 * This is a special case since stat/access is called on a file
		 * that does not exist (no inode value to process). The race
		 * condition is handled by failing the open call in step 3.
		 * It is impossible to stop the race condiiton in step 2 since
		 * the name used by an attacker won't necessarily match the
		 * name used in step 1 (i.e. /tmp/race vs /tmp/../tmp/race)
		 * and there are no inodes values to compare.
		 */

		/* only failed stat, lstat & access call entries are placed on
			stat_hash */

		for (tmp = HASH_FIRST(stat, pid); tmp != NULL; tmp = tmp->next)
		{
			if (pid == tmp->pid && strcmp(tmp->name, name) == 0)
			{
				lock_release();
				return(RP_PROCESS_FAIL_CALL);
			}
		}


		/*
		 * special case 2:
		 *
		 * unlink followed by a write:
		 *	Victim:		deletes a file
		 *	Attacker:	creates a link with the same name
		 *	Victim:		open(O_CREAT, name) opens the existing
		 * 			file and trashes its contents
		 *
		 *
		 * This is a special case since write/creat open call is called
		 * on a file that we don't know the inode value of. It is
		 * impossible to stop the race condiiton in step 2 since the
		 * name used by an attacker won't necessarily match the name
		 * used in step 1 (i.e. /tmp/race vs /tmp/../tmp/race) and
		 * there are no inode values to compare.
		 */

		/* only recently deleted file names are placed on unlink_hash */

		for (tmp = HASH_FIRST(unlink, pid); tmp != NULL; tmp= tmp->next)
		{
			if (pid == tmp->pid && strcmp(tmp->name, name) == 0)
			{
				lock_release();
				return(RP_PROCESS_FAIL_CALL);
			}
		}
	}


	/*
	 * Normal case
	 *
	 * check the operation on the current inode against all the previous
	 * operations involving the same inode
	 */

	for (tmp = HASH_FIRST(rp, inode); tmp != NULL; tmp = tmp->next)
	{
		if (tmp->inode != inode || tmp->fsid != fsid)
			continue;

		found = 1;


		if (tmp->pid == pid)
		{
			/* replace the current entry with the new operation */

			tmp->operation = operation;
			tmp->tick = ticks;

			/* we have checked the new operation against all
			   the preceeding operations until the current one
			   (they are sorted by inserted time). Since no
			   violations were detected until this point, we can
			   safely exit.
			*/

			lock_release();
			return(RP_PROCESS_OK);
		}


		if (tmp->operation == UNLINK)
		{
			/*
			 * inode collision:
			 * since the previous operation is unlink,
			 * the only way we can reach this code is if the
			 * inode is reused by another file
			 */

			//continue;
			//uprintf("inode collision %d %s\n", tmp->inode, tmp->name);

			lock_release();
			return(RP_PROCESS_OK);
		}


		if (deny_all)
		{
			if (deny_all_policy(tmp->operation, operation) ==
								POLICY_OK)
			{
				lock_release();
				return(RP_PROCESS_OK);
			}
		}
		else
		{
			if (permit_all_policy(tmp->operation, operation) ==
								POLICY_NOT_OK)
			{
				lock_release();
				return(RP_PROCESS_RESPOND);
			}
		}

		if (deny_all)
			log(9, "raceprot: Unknown operation sequence: original_pid=%d current_pid=%d inode=%d (%s) original_operation=%s current_operation=%s\n", tmp->pid, pid, inode, tmp->name, fileop_name(tmp->operation), fileop_name(operation));
	}



	/* found at least one matching inode entry, not our pid though.
	   possible interference. */

	if (found)
	{
		lock_release();

		/* deny all policy denies all requests by default */
		if (deny_all)
			return(RP_PROCESS_RESPOND);
		else
			return(RP_PROCESS_OK);
	}


	lock_release();

	return(RP_PROCESS_INSERT);
}



/*
 * deny_all_policy:	Deny All Ruleset Implementation
 *
 *			make sure the current operation doesn't interfere with
 *			any ongoing operations involving the same inode.
 *			reject any sequences of operations that are not known
 *			to be valid.
 *
 * input:		previous_op	previously observed operation involving
 *					the same inode as current_op operation
 *			current_op	current inode operation
 *
 * output:		POLICY_OK	no race conditions found
 *			POLICY_NOT_OK	race condition found
 */

int
deny_all_policy(int previous_op, int current_op)
{

#include "deny.policy"

	return(POLICY_NOT_OK);
}



/*
 * permit_all_policy:	Permit All Ruleset Implementation
 *
 *			make sure the current operation doesn't interfere with
 *			any ongoing operations involving the same inode.
 *			reject any sequences of operations that are known to
 *			be bad.
 *
 * input:		previous_op	previously observed operation involving
 *					the same inode as current_op operation
 *			current_op	current inode operation
 *
 * output:		POLICY_OK	no race conditions found
 *			POLICY_NOT_OK	race condition found
 */

int
permit_all_policy(int previous_op, int current_op)
{

#include "permit.policy"

	return(POLICY_OK);
}



/*
 * rp_register_fd:	register a file descriptor with an existing entry
 *			(only used by sys_open)
 *
 * input:		pid & inode of the current process & operation
 *			fd file descriptor to register
 *
 * output:		1 in case of success
 *			0 in case of failure (lock failed, no matching entry found)
 */

#ifdef MEDIATE_SYS_CLOSE

int
rp_register_fd(pid_t pid, int inode, int fd)
{
	struct rp_entry	*tmp;


	if (lock_shared() != 0)
	{
		uprintf("rp_register_fd(): failed to obtain a shared lock\n");
		return(0);
	}


	for (tmp = HASH_FIRST(rp, inode); tmp != NULL; tmp = tmp->next)
	{
		if (tmp->inode == inode && tmp->pid == pid && tmp->fsid == fsid)
		{
			/* only open operations can have a file descriptor
				associated with them */

			if (tmp->operation & OPEN)
			{
				tmp->fd = fd;
			}

			lock_release();

			return(1);
		}
	}

	lock_release();

	return(0);
}

#endif


/*
 * rp_insert_nolock:	creates a new entry
 *
 * input:	pid	pid of the process invoking the operation
 *		inode	inode involved in the operation
 *		fsid	.
 *		name	[file]name associated with the inode
 *		operation operation being carried on the inode
 *		op_tick	current timestamp
 *		op_type	operation type:
 *			RP_OP_NORMAL	all entries
 *			RP_OP_STAT	failed stat()/lstat()/access() entries
 *			RP_OP_UNLINK	unlink()'ed entries
 *
 * output:	1 in case of success
 *		0 in case of failure (out of memory)
 *
 * Note: rp_insert_nolock() expects exclusive lock to be acquired already
 */

int
rp_insert_nolock(pid_t pid, int inode, int fsid, char *name, enum FILEOP operation, int op_tick, int op_type)
{
	struct rp_entry	*new, *tmp;


	if (op_type == RP_OP_NORMAL)
	{
		/* find the last element */

		for (tmp = HASH_FIRST(rp, inode);
			tmp != NULL && tmp->next != NULL; tmp = tmp->next)
		    ;
	}
	else if (op_type == RP_OP_STAT)
	{
		for (tmp = HASH_FIRST(stat, pid);
			tmp != NULL && tmp->next != NULL; tmp = tmp->next)
		{
			if (tmp->pid == pid && strcmp(tmp->name, name) == 0)
			{
				/* do not insert duplicates */
				return(1);
			}
		}
	}
	else if (op_type == RP_OP_UNLINK)
	{
		for (tmp = HASH_FIRST(unlink, pid);
			tmp != NULL && tmp->next != NULL; tmp = tmp->next)
		{
			if (tmp->pid == pid && strcmp(tmp->name, name) == 0)
			{
				/* do not insert duplicates */
				return(1);
			}
		}
	}
	else
	{
		uprintf("rp_insert: Unknown operation type\n");
		return(0);
	}


	new = rp_malloc(strlen(name)+1);

	if (new == NULL)
		return(ENOMEM);

	strlcpy(new->name, name, strlen(name) + 1);

	new->next = NULL;
	new->tick = op_tick;
	new->pid = pid;
	new->inode = inode;
	new->fsid = fsid;
	new->operation = operation;
	new->fd = 0;


	if (tmp == NULL)
	{
		/* fill in the first entry of the hash chain */

		if (op_type == RP_OP_NORMAL)
		{
			HASH_FIRST(rp, inode) = new;
		}
		else if (op_type == RP_OP_STAT)
		{
			HASH_FIRST(stat, pid) = new;
		}
		else if (op_type == RP_OP_UNLINK)
		{
			HASH_FIRST(unlink, pid) = new;
		}
	}
	else
	{
		tmp->next = new;
	}

	return(1);
}



/*
 * rp_insert:		rp_insert_nolock() wrapper
 *			obtain an exclusive lock and call rp_insert_nolock()
 *
 * input/output:	see rp_insert_nolock()
 */

inline int
rp_insert(pid_t pid, int inode, int fsid, char *name, enum FILEOP operation, int op_tick, int op_type)
{
	int	ret;

	if (lock_exclusive() != 0)
	{
		uprintf("rp_insert: failed to obtain an exclusive lock\n");
		return(0);
	}

	ret = rp_insert_nolock(pid, inode, fsid, name, operation, op_tick, op_type);

	lock_release();

	return(ret);
}



/*
 * rp_malloc:	allocate the memory for the new entry (structure + filename)
 *
 * input:	size of the filename
 *
 * output:	pointer to the allocated structure
 *
 * Note:	needs exclusive access to modify free_list
 *		(provided by rp_insert lock)
 */

inline struct rp_entry*
rp_malloc(int size)
{
	struct rp_entry	*new;


	/*
	 * call malloc only when
	 * - free_list is empty or
	 * - size of the filename is over 31 characters long
	 *   (we don't cache entries with long filenames)
	 * otherwise return an entry from a free_list
	 */

	if (free_list_num == 0 || size > 31)
	{
		new = (struct rp_entry*) malloc(sizeof(struct rp_entry), M_FREE, M_NOWAIT);

		if (new == NULL)
		{
			uprintf("new = NULL in rp_malloc()\n");
			return(NULL);
		}

		new->name = (char*) malloc(size > 31 ? size : 32, M_FREE, M_NOWAIT);

		if (new->name == NULL)
		{
			uprintf("new->name = NULL in rp_malloc()\n");
			free(new, M_FREE);
			return(NULL);
		}
	}
	else
	{
		/* return the first entry off of a free_list */

		new = free_list.next;
		free_list.next = free_list.next->next;

		--free_list_num;
	}

	return(new);
}



/*
 * rp_free:	add the entry to a free_list unless the free_list is full
 *		already, in that case free() the entry
 *		
 * input:	ptr the entry to deallocate
 *
 * output:	none
 *
 * Note:	needs exclusive access to modify free_list
 *	 	(provided by rp_purge() lock)
 */

inline void
rp_free(struct rp_entry *ptr)
{
	/*
	 * do not cache the entries if
	 *	- free list is full or
	 *	- size of the string is over 31 bytes or
	 *	- we are unloading the module
	 */

	if ((free_list_num > FREE_LIST_NUM) || strlen(ptr->name) > 31 || unloading_module)
	{
		free(ptr->name, M_FREE);
		free(ptr, M_FREE);
	}
	else
	{
		/* append ptr to the free_list */

		ptr->next = free_list.next;
		free_list.next = ptr;

		++free_list_num;
	}
}



/*
 * rp_purge:	remove the stale (timed out) entries or
 *		all the entries if we are unloading the module
 *
 * input:	type	purge operation type
 *			PURGE_TYPE_TIMEOUT	timeout
 *			PURGE_TYPE_PID		purge by pid
 *			PURGE_TYPE_FD		purge by fd
 *
 *		pid	pid of the entries to remove
 *		(pid is set to 0 when we are removing in case of a timeout)
 *
 *		fd file descriptor of the entries to remove
 *		(fd is set to 0 when we are removing by pid or a timeout)
 *
 * output:	none
 */


void
rp_purge(int type, int pid, int fd)
{
	struct rp_entry	*ptr, *next, *prev, *tmp;
	int		i;


	if (lock_exclusive() != 0)
	{
		uprintf("rp_purge: failed to obtain an exclusive lock\n");
		return;
	}


#define	RP_FREE()					\
	tmp = ptr;					\
	if (prev != NULL)				\
		prev->next = next;			\
	else						\
		rp_hash[i] = ptr = next;		\
	rp_free(tmp);


	for (i = 0; i < rp_HASH_SIZE; i++)
	{
		prev = NULL;
		ptr = rp_hash[i];

		for (; ptr != NULL; )
		{
			next = ptr->next;

			if (type & PURGE_TYPE_TIMEOUT)
			{
				if (((ticks - ptr->tick) > timeout_purge) ||
				    unloading_module)
				{
					RP_FREE();
				}
				else
				{
					/*
					 * if purging by timeout, we can bail
					 * out earlier by giving up on the first
					 * entry that fails the time comparison
					 * since the entries are sorted by
					 * inserted time
					 */

					break;
				}
			}
#ifdef MEDIATE_SYS_CLOSE
			else if ((type & PURGE_TYPE_FD) &&
					ptr->pid == pid && ptr->fd == fd)
			{
				/* if we are purging by a file descriptor
				   then bail after the first match since there
				   can by only one unique descriptor
				   associated with each pid */

				RP_FREE();

				lock_release();
				return;
			}
#endif
			else if (((type & PURGE_TYPE_PID) && ptr->pid == pid))
			{
				RP_FREE();
			}
			else
			{
				prev = ptr;
			}

			ptr = next;
		}
	}


	/* clean the stat hash */

	for (i = 0; i < stat_HASH_SIZE; i++)
	{
		prev = NULL;
		ptr = stat_hash[i];

		for (; ptr != NULL; )
		{
			next = ptr->next;

			if ((ticks - ptr->tick) > timeout_purge)
			{
				tmp = ptr;	

				if (prev != NULL)
					prev->next = next;
				else
					stat_hash[i] = ptr = next;
			
				rp_free(tmp);
			}
			else
			{
				prev = ptr;
			}

			ptr = next;
		}
	}


	/* clean the unlink hash */

	for (i = 0; i < unlink_HASH_SIZE; i++)
	{
		prev = NULL;
		ptr = unlink_hash[i];

		for (; ptr != NULL; )
		{
			next = ptr->next;

			if ((ticks - ptr->tick) > timeout_purge)
			{
				tmp = ptr;	

				if (prev != NULL)
					prev->next = next;
				else
					unlink_hash[i] = ptr = next;
			
				rp_free(tmp);
			}
			else
			{
				prev = ptr;
			}

			ptr = next;
		}
	}


	lock_release();
}



/*
 * rp_duplicate_entry:	duplicate all the entries associated with a pid
 *			(called by sys_*fork)
 *
 * input:		oldpid pid of the entries to duplicate
 *			newpid new pid to associate all entries with
 *
 * output:		none
 */

void
rp_duplicate_entry(int oldpid, int newpid)
{
	struct rp_entry	*ptr;
	int		i;


	if (lock_exclusive() != 0)
	{
		uprintf("rp_duplicate_entry: failed to obtain an exclusive lock\n");
		return;
	}


	for (i = 0; i < rp_HASH_SIZE; i++)
	{
		for (ptr = rp_hash[i]; ptr != NULL; ptr = ptr->next)
		{
			if (ptr->pid == oldpid)
			{
				rp_insert_nolock(newpid, ptr->inode, ptr->fsid,
						 ptr->name, ptr->operation,
						 ticks, RP_OP_NORMAL);
			}
		}
	}


	lock_release();
}



/*
 * rp_timeout:	soft timer callout (called every TIMEOUT_TIMER/1000 seconds)
 *		call rp_purge to purge the stale entries
 *
 * input:	arg timeout_add parameter
 *
 * output:	none
 */

void
rp_timeout(void *arg)
{
	purge_needed = 1;

	timeout_add((struct timeout*)arg, TIMEOUT_TIMER);
}



/*
 * rp_init:	Initialization function called upon module loading.
 *		Initializes the locks, timeouts and various data structures
 *
 * input:	none
 *
 * output:	none
 */

void
rp_init()
{
	bzero(&rp_hash, sizeof(rp_hash));
	bzero(&stat_hash, sizeof(stat_hash));
	bzero(&unlink_hash, sizeof(unlink_hash));

	unloading_module = 0;

	timeout_set(&to, rp_timeout, &to);
	timeout_add(&to, TIMEOUT_TIMER);

	lockinit(&rp_lock, PCATCH, "tm", 0, 0);
}



/*
 * rp_fini:	Called upon module unloading.
 *		Removes the timeout & frees the data structures
 *
 * input:	none
 *
 * output:	none
 */

void
rp_fini()
{
	struct rp_entry	*ptr, *tmp;
	int		i;


	unloading_module = 1;

	timeout_del(&to);

	rp_purge(PURGE_TYPE_TIMEOUT, 0, 0);


	/* free all the entries on the free_list */

	for (ptr = free_list.next; ptr != NULL; )
	{
		tmp = ptr;
		ptr = ptr->next;

		free(tmp->name, M_FREE);
		free(tmp, M_FREE);
	}


	/* free all the stat hash entries */

	for (i = 0; i < stat_HASH_SIZE; i++)
	{
		for (ptr = stat_hash[i]; ptr != NULL; )
		{
			tmp = ptr;
			ptr = ptr->next;

			free(tmp->name, M_FREE);
			free(tmp, M_FREE);
		}
	}


	/* free all the unlink hash entries */

	for (i = 0; i < unlink_HASH_SIZE; i++)
	{
		for (ptr = unlink_hash[i]; ptr != NULL; )
		{
			tmp = ptr;
			ptr = ptr->next;

			free(tmp->name, M_FREE);
			free(tmp, M_FREE);
		}
	}
}


